import numpy as np
from item_models import Model, PL2


def estimate_ability_max_lik(
    params: np.ndarray,
    results: np.ndarray,
    end=0.00000001,
    eps=0.01,
    eps2=0.000001,
) -> np.ndarray:
    """
    Estimate abilities using maximum likelihood estimation.

    Parameters
    ----------
        params : np.ndarray
            Item parameters. Each column is a different item, each row is
            a different parameter.
        results : np.ndarray
            Results. Each row is a different individual, each column is a
            different item.
        end : float
            Convergence threshold.
        eps : float
            Threshold for detecting a near-singular matrix.
        eps2 : float
            Threshold for detecting small values of P*Q.

    Returns
    -------
        np.ndarray
            Array of abilities.
    """
    params = np.array(params)
    results = np.array(results)
    est = 0.5
    prev_est = 0
    while abs(est - prev_est) > end:
        P = PL2.pl2_p(est, params)[0]
        W = (1 - P) * P
        denom = np.sum(params[1] ** 2 * W)
        if abs(denom) < eps or np.any(np.abs(W) < eps2):
            print('break1')
            break
        prev_est = est
        est = est + (np.sum(params[1] * W * ((results - P) / W)) / denom)
    return est


def estimate_item_params_max_lik(
    model: Model,
    ability: np.ndarray,
    result: np.ndarray,
    a_orig=1,
    b_orig=0,
    end=1e-10,
    eps=1e-10,
) -> np.ndarray:
    """
    Estimate item parameters using maximum likelihood estimation.

    Parameters
    ----------
        model : Model
            Instance of PL2 model.
        ability : np.ndarray
            Array of abilities.
        result : np.ndarray
            Results. Each row is a different individual, each column is a
            different item.
        sigm_orig : float
            Initial value for the discrimination parameter.
        lamb_orig : float
            Initial value for the difficulty parameter.
        end : float
            Convergence threshold.
        eps : float
            Threshold for detecting a near-singular matrix.

    Returns
    -------
        np.ndarray
            Array of item parameters.
    """
    ability = np.array(ability)
    result = np.array(result)
    est = np.array([a_orig, b_orig])
    prev_est = est + 2 * end
    while np.all(np.abs(est - prev_est) > end):
        P = PL2.pl2_p(ability, [[est[0]], [est[1]]]).transpose()[0]
        W = P * (1 - P)
        lamb11 = -np.sum((ability - est[1]) ** 2 * W)
        lamb12 = est[0] * np.sum((ability - est[1]) * W)
        lamb22 = -np.sum(est[0] ** 2 * W)
        L = np.array([[lamb11, lamb12], [lamb12, lamb22]])
        if abs(np.linalg.det(L)) < eps:
            L += np.identity(2) * 1e-10
        L_inv = np.linalg.inv(L)

        L1 = np.sum((result - P) * (ability - est[1]))
        L2 = np.sum((P - result) * est[0])
        obs_mat = np.array([L1, L2])
        prev_est = est
        est = est - np.matmul(L_inv, obs_mat)
        
        # enforce bounds
        for par in range(model.num_params):
            est[par] = model.param_bounds[par][0] if est[par] < model.param_bounds[par][0] else est[par]
            est[par] = model.param_bounds[par][1] if est[par] > model.param_bounds[par][1] else est[par]
    return est



from pathlib import Path
import sys

if str(Path(__file__).parent) not in sys.path:
    sys.path.append(str(Path(__file__).parent))
from item_models import Model, PL1, PL2, PL3, Norm
import numpy as np
import jmle.pl1 as pl1
import jmle.pl2 as pl2
import jmle.pl3 as pl3


def estimate_ability_max_lik(
    model: Model, params: np.ndarray, results: np.ndarray, **kwargs: dict
) -> np.ndarray:
    if isinstance(model, PL1):
        return pl1.estimate_ability_max_lik(params, results, **kwargs)
    if isinstance(model, PL2):
        return pl2.estimate_ability_max_lik(params, results, **kwargs)
    if isinstance(model, PL3):
        return pl3.estimate_ability_max_lik(params, results, **kwargs)
    raise NotImplementedError(
        "Cannot estimate ability using maximum liklihood estimation for given model."
    )


def estimate_item_params_max_lik(
    model: Model,
    ability: np.ndarray,
    result: np.ndarray,
    **kwargs,
) -> np.ndarray:
    if isinstance(model, PL1):
        return pl1.estimate_item_params_max_lik(model, ability, result, **kwargs)
    if isinstance(model, PL2):
        return pl2.estimate_item_params_max_lik(model, ability, result, **kwargs)    
    if isinstance(model, PL3):
        return pl3.estimate_item_params_max_lik(model, ability, result, **kwargs)
    raise NotImplementedError(
        "Cannot estimate parameters using maximum liklihood estimation for given model."
    )


def em(
    model: Model,
    responses: np.ndarray,
    eps: float = 0.1,
    max_iterations: int = 1000,
) -> tuple:
    """
    Estimate abilities and item parameters using the EM algorithm with
    joint maximum liklihood estimation.

    Parameters
    ----------
        responses : np.ndarray
            Responses to items. Each row is a different subject, each
            column is a different item.
        eps : float
            Convergence threshold.
        max_iterations : int, optional
            Maximum number of iterations. Defaults to 1000.

    Returns
    -------
        np.ndarray
            Estimated abilities
        np.ndarray
            Estimated item parameters. Each column is a different item,
            each row is a different parameter.
    """
    responses = np.array(responses)
    subjects, items = responses.shape
    abilities = np.zeros(subjects)
    prev_abilities = abilities + 2 * eps
    iteration = 0
    while (
        np.any(np.abs(abilities - prev_abilities) > eps) and iteration < max_iterations
    ):
        print(1)
        params = np.array(
            [
                estimate_item_params_max_lik(model, abilities, responses[:, i])
                for i in range(items)
            ]
        ).T
        prev_abilities = abilities
        print(2)
        # abilities = np.array(
        #     [
        #         estimate_ability_max_lik(model, params, responses[i])
        #         for i in range(subjects)
        #     ]
        # )
        abilities = []
        for i in range(subjects):
            print(model, params, responses[i])
            abilities.append(estimate_ability_max_lik(model, params, responses[i]))
        abilities = np.array(abilities)
        iteration += 1
    return abilities, params



import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
import scipy.integrate as integrate
from scipy.optimize import minimize

class Generic():
    def __init__(self, model) -> None:
        self.p = model.p
        self.num_params = model.num_params
        self.ability_prior = lambda ability: norm.pdf(ability)

    # eq 6.2
    def resp_prob(self, item_responses, ability, params):
        ''' Probability of the subject's item response vector conditional 
            on the examinee's ability and item parameters'''
        P = self.p(ability, params)
        Q = 1 - P
        return np.prod(P ** item_responses * Q ** (1 - item_responses), axis=1)

    # integrand of denominator of eq 6.1
    def marg_resp_prob_integrand(self, item_responses, ability, params):
        return self.ability_prior(ability) * self.resp_prob(item_responses, ability, params)

    # denominator of eq 6.1
    # consider other methods of integration, and how the user can specify the integration parameters
    def marg_resp_prob(self, item_responses, params):
        ''' Marginal probability of the item response vector with respect 
            to the item parameters and the population ability density, i.e. 
            integrate marg_resp_prob_integrand w.r.t. ability'''
        integrand = lambda x: self.marg_resp_prob_integrand(item_responses, x, params)
        result, _ = integrate.quad(integrand, -6, 6, epsabs=1e-1, epsrel=1e-1)
        return result

    # Eq 6.4. log likelihood func to be maximized
    def marg_log_liklihood(self, data, params):
        ret = 0
        for item_response in data:
            ret += np.log(self.marg_resp_prob(item_response, params))
        return ret

    # get param value that maximizes marg_log_liklihood
    # consider other methods of optimization, and how the user can specify the optimization parameters
    # show error if optimization failed
    def maximize_log_likelihood_for_param(self, data, params, item_idx, param_idx, bounds):
        initial_guess = params[param_idx][item_idx]
        params = params.copy()
        def objective_func(x):
            params[param_idx][item_idx] = x
            return -self.marg_log_liklihood(data, params)
        result = minimize(objective_func, initial_guess, bounds=[bounds], tol=1e-1, method='Nelder-Mead')
        return result.x[0], -result.fun

    def em1(self, data, prov_params, bounds, eps=1e-6):
        data = np.array(data)
        subjects, items = data.shape
        params = np.tile(np.array(prov_params, dtype=float), items).reshape(items, self.num_params).T
        
        prev_log_L = -np.inf
        log_L = self.marg_log_liklihood(data, params)
        while log_L - prev_log_L > eps:
            print(f'prev log L: {prev_log_L}')
            print(f'log L: {log_L}')
            prev_log_L = log_L
            new_params = np.zeros((self.num_params, items), dtype=float)
            for item_idx in range(items):
                print(f'Item:{item_idx}')
                for param_idx in range(self.num_params):
                    new_param, _ = self.maximize_log_likelihood_for_param(data, params, item_idx, param_idx, bounds[param_idx])
                    new_params[param_idx][item_idx] = new_param
            
            params = new_params
            log_L = self.marg_log_liklihood(data, params)
        return params

    # preferred # add params to constructor?
    def em2(self, data, prov_params, bounds, eps=1):
        data = np.array(data)
        subjects, items = data.shape
        params = np.tile(np.array(prov_params, dtype=float), items).reshape(items, self.num_params).T
        
        prev_log_L = -np.inf
        log_L = self.marg_log_liklihood(data, params)
        while log_L - prev_log_L > eps:
            print(f'prev log L: {prev_log_L}')
            print(f'log L: {log_L}')
            prev_log_L = log_L
            for item_idx in range(items):
                print(f'Item:{item_idx}')
                for param_idx in range(self.num_params):
                    new_param, _ = self.maximize_log_likelihood_for_param(data, params, item_idx, param_idx, bounds[param_idx])
                    params[param_idx][item_idx] = new_param
            log_L = self.marg_log_liklihood(data, params) # unnecessary, returned by maximize_log_likelihood_for_param
        return params

    def em(self, data, prov_params, bounds, eps=1e-6):
        data = np.array(data)
        _, items = data.shape
        params = np.tile(np.array(prov_params, dtype=float), items).reshape(items, self.num_params).T
        
        prev_log_L = -np.inf
        log_L = self.marg_log_liklihood(data, params)
        while log_L - prev_log_L > eps:
            print(f'prev log L: {prev_log_L}')
            print(f'log L: {log_L}')
            prev_log_L = log_L
            new_params = np.zeros((self.num_params, items), dtype=float)
            for item_idx in range(items):
                print(f'Item:{item_idx}')
                new_params = np.zeros(self.num_params, dtype=float)
                for param_idx in range(self.num_params):
                    new_param, _ = self.maximize_log_likelihood_for_param(data, params, item_idx, param_idx, bounds[param_idx])
                    new_params[param_idx] = new_param
                
                for param_idx in range(self.num_params):
                    params[param_idx][item_idx] = new_params[param_idx]
                    
            log_L = self.marg_log_liklihood(data, params)
        


import numpy as np
from item_models import Model, PL3


def estimate_ability_max_lik(
    params: np.ndarray,
    results: np.ndarray,
    end=1e-10,
    eps=1e-10,
    bounds=(-7, 7),
) -> np.ndarray:
    """
    Estimate abilities using maximum likelihood estimation.
    """
    params = np.array(params)
    results = np.array(results)
    
    # def log_lik_der(est):
    #     P = PL3.pl3_p(est, params)[0]
    #     Q = 1 - P
    #     P_2pm = PL3.pl2_p(est, params)[0]
    #     W = P * Q
    #     if np.any(np.abs(W) < eps):
    #         W += 1e-10
    #     # if np.any(np.abs(P) < eps):
    #     #     P += 1e-10
    #     return np.sum(params[0] * W * ((results - P) / W) * (P_2pm / P))
    
    # def log_lik(est):
    #     P = PL3.pl3_p(est, params)[0]
    #     Q = 1 - P
    #     P_2pm = PL3.pl2_p(est, params)[0]
    #     W = P * Q
    #     if np.any(np.abs(P) < eps):
    #         P += 1e-10
    #     if np.any(np.abs(Q) < eps):
    #         Q += 1e-10
    #     return np.sum(np.log(P) * results + np.log(Q) * (1 - results))
        
    # ability = np.linspace(-10, 10)
    # der = np.array([log_lik_der(a) for a in ability])
    # ll = np.array([log_lik(a) for a in ability])
    # import matplotlib.pyplot as plt
    # plt.plot(ability, der)
    # plt.plot(ability, ll)
    # plt.show()
    # quit()
    
    est = 0
    prev_est = est + 2 * end
    i = 0
    while abs(est - prev_est) > end and i < 15:
        P = PL3.pl3_p(est, params)[0]
        Q = 1 - P
        
        P_2pm = PL3.pl2_p(est, params)[0]
        W = P * Q
        denom = -np.sum(params[0] ** 2 * W * (P_2pm / P) ** 2)
        if abs(denom) < eps:
            print(f"denom: {denom}")
            denom += 1e-10

        if np.any(np.abs(W) < eps):
            print(f"W: {W}")
            W += 1e-10

        num = np.sum(params[0] * W * ((results - P) / W) * (P_2pm / P))
        prev_est = est
        est = est - (num / denom)
        print(est)
        est = bounds[0] if est < bounds[0] else est
        est = bounds[1] if est > bounds[1] else est
        print(est)
        print()
        i+=1
    return est


def estimate_item_params_max_lik(
    model: Model,
    ability: np.ndarray,
    result: np.ndarray,
    a_orig=1,
    b_orig=0,
    c_orig=0.1,
    end=1e-10,
    eps=1e-10,
) -> np.ndarray:
    """
    Estimate item parameters using maximum likelihood estimation.

    Parameters
    ----------
        model : Model
            Instance of PL3 model.
        ability : np.ndarray
            Array of abilities.
        result : np.ndarray
            Results. Each row is a different individual, each column is a
            different item.
        sigm_orig : float
            Initial value for the discrimination parameter.
        lamb_orig : float
            Initial value for the difficulty parameter.
        end : float
            Convergence threshold.
        eps : float
            Threshold for detecting a near-singular matrix.

    Returns
    -------
        np.ndarray
            Array of item parameters.
    """
    ability = np.array(ability)
    result = np.array(result)
    est = np.array([a_orig, b_orig, c_orig], dtype=float)
    prev_est = est + 2 * end
    while np.all(np.abs(prev_est - est) > end):
        P = PL3.pl3_p(ability, est.reshape((3, 1))).T[0]
        if np.any(np.abs(P - est[2]) < eps) or np.any(np.abs(P) < eps):
            P = P + 1e-10
        Q = 1 - P
        P_2pm = PL3.pl2_p(ability, est.reshape((3, 1))).T[0]

        L11 = -np.sum((ability - est[1]) ** 2 * P * Q * (P_2pm / P) ** 2)
        L12 = np.sum(est[0] * (ability - est[1]) * P * Q * (P_2pm / P))
        L13 = -np.sum((ability - est[1]) * (Q / (1 - est[2])) * (P_2pm / P))
        L22 = -est[0] ** 2 * np.sum(P * Q * (P_2pm / P))
        L23 = np.sum(est[0] * (Q / (1 - est[2])) * (P_2pm / P))
        L33 = -np.sum((Q / (1 - est[2])) / (P - est[2]) * (P_2pm / P))
        L = np.array([[L11, L12, L13], [L12, L22, L23], [L13, L23, L33]])

        if abs(np.linalg.det(L)) < eps:
            L += np.identity(3) * 1e-10
        L_inv = np.linalg.inv(L)

        L1 = np.sum((result - P) * (ability - est[1]) * (P_2pm / P))
        L2 = -est[0] * np.sum((result - P) * (P_2pm / P))
        L3 = np.sum((result - P) / (P - est[2]) * (P_2pm / P))
        obs_mat = np.array([L1, L2, L3])
        prev_est = est
        est = est - np.matmul(L_inv, obs_mat)

        # enforce bounds
        for par in range(model.num_params):
            est[par] = (
                model.param_bounds[par][0]
                if est[par] < model.param_bounds[par][0]
                else est[par]
            )
            est[par] = (
                model.param_bounds[par][1]
                if est[par] > model.param_bounds[par][1]
                else est[par]
            )
    return est



from model import Model
import numpy as np
from scipy.stats import norm
from typing import Union


class Norm(Model):
    """
    Normal ogive model.
    """

    def __init__(self):
        super().__init__(
            Norm.norm_p,
            num_params=2,
            multi=True,
            prov_params=[0, 1],
            loc_param=1,
            param_bounds=((-10, 10), (-4, 4)),
            loc_param_asc=True,
        )

    @staticmethod
    def norm_p(
        ability: np.ndarray,
        params: np.ndarray,
    ) -> np.ndarray:
        """
        Calculate the probability of a correct response given the ability and
        item parameters.

        Parameters
        ----------
            ability : np.ndarray
                Array of abilities.
            params : np.ndarray
                Item parameters. Item parameters. Each column is a different
                item, each row is a different parameter.

        Returns
        -------
            np.ndarray
                Array of probabilities. Each row is a different individual,
                each column is a different item.
        """
        ability = np.array(ability)
        params = np.array(params)
        z = np.matmul(ability[:, np.newaxis], [params[1]])
        z += np.tile(params[0], (len(ability), 1))
        return norm.cdf(z)

    @staticmethod
    def norm_density(self, ability: np.ndarray, params: np.ndarray) -> np.ndarray:
        """
        Normal density function.
        """
        ability = np.array(ability)
        params = np.array(params)
        z = np.matmul(ability[:, np.newaxis], [params[1]])
        z += np.tile(params[0], (len(ability), 1))
        return norm.pdf(z)

    @staticmethod
    def convert_param_form(self, params):
        return np.array([-params[0] / params[1], 1 / params[1]])



import pytest
from openirt import item_models
import numpy as np

# ---GENERIC---

def test_prob_gen():
    params = np.array([[0.45, 0.6], [0.1, 0.15]])
    abilities = np.array([0.55, 1.2])
    def p(ab, params):
        if ab > params[0]:
            return 1
        else:
            return params[1]
    model = item_models.Model(p, num_params=2)
    prob = model.p(abilities, params)
    assert prob.shape == (2,2)
    assert prob[0][0] == 1
    assert prob[0][1] == 0.15
        


from model import Model
import numpy as np
from pl2 import PL2


class PL1(Model):
    """
    1-Parameter logistic model (Rasch model).
    p(theta) = 1 / (1 + exp(-(zeta + theta))
    """

    def __init__(self):
        super().__init__(
            PL1.pl1_p,
            num_params=1,
            multi=True,
            prov_params=[0],
            loc_param=0,
            param_bounds=[(-4, 4)],
            loc_param_asc=True,
        )

    @staticmethod
    def pl1_p(ability: np.ndarray, params: np.ndarray) -> np.ndarray:
        """
        Probability of a correct response given the ability and item
        parameters.

        Parameters
        ----------
            ability : np.ndarray
                Array of abilities.
            params : np.ndarray
                Item parameters. Item parameters. Each column is a different
                item, each row is a different parameter.

        Returns
        -------
            np.ndarray
                Array of probabilities. Each row is a different individual,
                each column is a different item.
        """
        return PL2().pl2_p(ability, [np.ones(len(params[0])), params[0]])


import unittest
from openirt import item_models
import numpy as np


class TestPL3(unittest.TestCase):
    def test_prob_shape(self):
        params = np.array([[-1, 0, 1], [-1.5, 0, 1.5], [0.1, 0.1, 0.1]])
        abilities = np.array([-1.2, 1.2])
        pl3 = item_models.PL3()
        prob = pl3.p(abilities, params)
        assert prob.shape == (2, 3)

    def test_valid_prob(self):
        pl3 = item_models.PL3()
        params = np.array(
            [
                np.linspace(*pl3.param_bounds[0]),
                np.linspace(*pl3.param_bounds[1]),
                np.linspace(*pl3.param_bounds[2]),
            ]
        )
        abilities = np.array([np.linspace(-10, 10)])
        prob = pl3.p(abilities, params)
        assert np.all(prob >= 0)
        assert np.all(prob <= 1)

    def test_prob_value(self):
        params = np.array([[1, 0, -1], [-1.5, 0, 1.5], [0.1, 0.2, 0.3]])
        abilities = np.array([-10, 1.5, 10])
        pl3 = item_models.PL3()
        prob = pl3.p(abilities, params)
        assert prob[0][0] < 1e-3 + 0.1
        assert prob[2][0] > 1 - 1e-3

    def test_simulated_resp(self):
        params = np.array([[1, 0, -1], [-1.5, 0, 1.5], [0.1, 0.2, 0.3]])
        abilities = np.array([-10, 1.5])
        pl3 = item_models.PL1()
        resp = pl3.simulated_responses(abilities, params)
        assert resp.shape == (2, 3)
        assert np.all(resp >= 0)
        assert np.all(resp <= 1)



from model import Model
import numpy as np


class PL2(Model):
    """
    2-parameter logistic model (a/b form).
    p(theta) = 1 / (1 + exp(-a(theta - b))
    """

    def __init__(self):
        super().__init__(
            PL2.pl2_p,
            num_params=2,
            multi=True,
            prov_params=[1, 0],
            loc_param=1,
            param_bounds=((-10, 10), (-4, 4)),
            loc_param_asc=True,
        )

    @staticmethod
    def pl2_p(ability: np.ndarray, params: np.ndarray) -> np.ndarray:
        """
        Probability of a correct response given the ability and item
        parameters.

        Parameters
        ----------
            ability : np.ndarray
                Array of abilities.
            params : np.ndarray
                Item parameters. Item parameters. Each column is a different
                item, each row is a different parameter.

        Returns
        -------
            np.ndarray
                Array of probabilities. Each row is a different individual,
                each column is a different item.
        """
        items = np.array(params).shape[1]
        num_abilities = 1
        if np.array(ability).shape != ():
            num_abilities = np.array(ability).shape[0]
        ability = np.tile(np.array([ability]).transpose(), (1, items))

        params = np.tile(params, num_abilities).reshape((2, num_abilities, items))
        return 1 / (1 + np.exp(-params[0] * (ability - params[1])))





import numpy as np
from item_models import PL2


def estimate_ability_max_lik(
    params: np.ndarray,
    results: np.ndarray,
    end=0.00000001,
    eps=0.01,
    eps2=0.000001,
) -> np.ndarray:
    """
    Estimate abilities using maximum likelihood estimation.

    Parameters
    ----------
        params : np.ndarray
            Item parameters. Each column is a different item, each row is
            a different parameter.
        results : np.ndarray
            Results. Each row is a different individual, each column is a
            different item.
        end : float
            Convergence threshold.
        eps : float
            Threshold for detecting a near-singular matrix.
        eps2 : float
            Threshold for detecting small values of P*Q.

    Returns
    -------
        np.ndarray
            Array of abilities.
    """
    params = np.array(params)
    results = np.array(results)
    est = 0.5
    prev_est = 0
    while abs(est - prev_est) > end:
        P = PL2.pl2_p(est, params)[0]
        W = (1 - P) * P
        denom = np.sum(params[1] ** 2 * W)
        if abs(denom) < eps or np.any(np.abs(W) < eps2):
            print('break1')
            break
        prev_est = est
        est = est + (np.sum(params[1] * W * ((results - P) / W)) / denom)
    return est


def estimate_item_params_max_lik(
    ability: np.ndarray,
    result: np.ndarray,
    sigm_orig=-1,
    lamb_orig=1,
    end=0.00001,
    eps=0.0001,
) -> np.ndarray:
    """
    Estimate item parameters using maximum likelihood estimation.

    Parameters
    ----------
        ability : np.ndarray
            Array of abilities.
        result : np.ndarray
            Results. Each row is a different individual, each column is a
            different item.
        sigm_orig : float
            Initial value for the discrimination parameter.
        lamb_orig : float
            Initial value for the difficulty parameter.
        end : float
            Convergence threshold.
        eps : float
            Threshold for detecting a near-singular matrix.

    Returns
    -------
        np.ndarray
            Array of item parameters.
    """
    ability = np.array(ability)
    result = np.array(result)
    est = [sigm_orig, lamb_orig]
    prev_est = [0, 0]
    i = 0
    while abs(est[0] - prev_est[0]) > end or abs(est[1] - prev_est[1]) > end:
        P = PL2.pl2_p(ability, [[est[0]], [est[1]]]).transpose()[0]
        W = P * (1 - P)
        L11 = -np.sum(W)
        L12 = -np.sum(ability * W)
        L22 = -np.sum(ability**2 * W)
        L = np.array([[L11, L12], [L12, L22]])
        print(est)
        if abs(np.linalg.det(L)) < eps:
            # print(f'break2: {abs(np.linalg.det(L))}')
            # break
            L += np.identity(2) * 1e-6
        L_inv = np.linalg.inv(L)

        obs_mat = np.array([np.sum(result - P), np.sum((result - P) * ability)])
        prev_est = est
        est = est - np.matmul(L_inv, obs_mat)
        i += 1
        if i > 10000:
            break
    return est



import numpy as np
import pandas as pd

def rescale(df):
    """Assumes that data contains integer values in ascending order of latent trait.
    Transforms data such that each item response contains all values from 0 to max."""
    rescaled_df = pd.DataFrame()
    for column in df.columns:
        ranks = df[column].rank(method="dense").astype(int) - 1
        rescaled_df[column] = ranks
    return rescaled_df

def remove_single_value_columns(df):
    columns_to_drop = []
    for column in df.columns:
        if df[column].nunique() == 1:
            columns_to_drop.append(column)
    df.drop(columns=columns_to_drop, inplace=True)
    return df



from pathlib import Path
import sys
if str(Path(__file__).parent) not in sys.path:
    sys.path.append(str(Path(__file__).parent))
    
# from irt_instance import IRTInstance
import item_models
import mmle
import jmle
from clean import rescale, remove_single_value_columns


from pathlib import Path
import sys
if str(Path(__file__).parent) not in sys.path:
    sys.path.append(str(Path(__file__).parent))
from model import Model
from pl1 import PL1
from pl2 import PL2
from pl3 import PL3
from norm import Norm


import unittest
from openirt import item_models
import numpy as np


class TestPL1(unittest.TestCase):
    def test_prob_shape(self):
        params = np.array([[-1.5, 0, 1.5]])
        abilities = np.array([-1.2, 1.2])
        pl1 = item_models.PL1()
        prob = pl1.p(abilities, params)
        assert prob.shape == (2, 3)

    def test_valid_prob(self):
        pl1 = item_models.PL1()
        params = np.array([np.linspace(*pl1.param_bounds[0])])
        abilities = np.array([np.linspace(-10, 10)])
        prob = pl1.p(abilities, params)
        assert np.all(prob >= 0)
        assert np.all(prob <= 1)

    def test_prob_value(self):
        params = np.array([[1.5, 0]])
        abilities = np.array([-10, 1.5, 10])
        pl1 = item_models.PL1()
        prob = pl1.p(abilities, params)
        assert prob[0][0] < 1e-3
        assert abs(prob[1][0] - 0.5) < 1e-3
        assert prob[2][0] > 1 - 1e-3

    def test_simulated_resp(self):
        params = np.array([[-1.5, 0, 1.5]])
        abilities = np.array([-1.2, 1.2])
        pl1 = item_models.PL1()
        resp = pl1.simulated_responses(abilities, params)
        assert resp.shape == (2, 3)
        assert np.all(resp >= 0)
        assert np.all(resp <= 1)

    def test_plot_icc(self):
        import matplotlib
        matplotlib.use('Agg')  # Set the backend to Agg before importing pyplot
        import matplotlib.pyplot as plt
        params = np.array([[-1.5, 0, 1.5]])
        pl1 = item_models.PL1()
        pl1.plot_icc(params)


import numpy as np
from item_models import Model, PL1
from jmle.pl2 import estimate_ability_max_lik as pl2_estimate_ability_max_lik


def estimate_ability_max_lik(
    params: np.ndarray,
    results: np.ndarray,
    end=0.00000001,
    eps=0.01,
    eps2=0.000001,
) -> np.ndarray:
    """
    Estimate abilities using the maximum likelihood method.
    See Model.estimate_ability_max_lik and PL2.estimate_ability_max_lik.
    """
    return pl2_estimate_ability_max_lik(
        [params[0], np.ones(len(params[0]))], results, end, eps, eps2
    )


def estimate_item_params_max_lik(
    model: Model,
    ability: np.ndarray,
    result: np.ndarray,
    sigm_orig=0,
    end=1e-10,
    eps=1e-10,
) -> np.ndarray:
    """
    Estimate item parameters using maximum likelihood estimation.

    Parameters
    ----------
        model : Model
            Instance of PL1 model.
        ability : np.ndarray
            Array of abilities.
        result : np.ndarray
            Results. Each row is a different individual, each column is a
            different item.
        sigm_orig : float
            Initial value for the discrimination parameter.
        lamb_orig : float
            Initial value for the difficulty parameter.
        end : float
            Convergence threshold.
        eps : float
            Threshold for detecting a near-singular matrix.

    Returns
    -------
        np.ndarray
            Array of item parameters.
    """
    ability = np.array(ability)
    result = np.array(result)
    est = sigm_orig
    prev_est = est + 2 * end
    while abs(est - prev_est) > end:
        P = PL1.pl1_p(ability, [[est]]).transpose()[0]
        L1 = np.sum(P - result)
        L2 = -np.sum(P * (1 - P))
        if abs(L2) < eps:
            L += 1e-10
        prev_est = est
        est = est - L1 / L2
        
        # enforce bounds
        est = model.param_bounds[0][0] if est < model.param_bounds[0][0] else est
        est = model.param_bounds[0][1] if est > model.param_bounds[0][1] else est
    return np.array([est])



import unittest
from openirt import item_models
import numpy as np


# class TestNorm(unittest.TestCase):
#     def test_prob_shape(self):
#         params = np.array([[-1, 0, 1], [-1.5, 0, 1.5]])
#         abilities = np.array([-1.2, 1.2])
#         norm = item_models.Norm()
#         prob = norm.p(abilities, params)
#         assert prob.shape == (2, 3)

#     def test_valid_prob(self):
#         pl2 = item_models.PL2()
#         params = np.array(
#             [np.linspace(*pl2.param_bounds[0]), np.linspace(*pl2.param_bounds[1])]
#         )
#         abilities = np.array([np.linspace(-10, 10)])
#         prob = pl2.p(abilities, params)
#         assert np.all(prob >= 0)
#         assert np.all(prob <= 1)

#     def test_prob_value(self):
#         params = np.array([[1, 0, -1], [-1.5, 0, 1.5]])
#         abilities = np.array([-10, 1.5, 10])
#         pl2 = item_models.PL2()
#         prob = pl2.p(abilities, params)
#         assert prob[0][0] < 1e-3
#         assert abs(prob[1][2] - 0.5) < 1e-3
#         assert prob[2][0] > 1 - 1e-3

#     def test_simulated_resp(self):
#         params = np.array([[1, 0, -1], [-1.5, 0, 1.5]])
#         abilities = np.array([-1.2, 1.2])
#         pl2 = item_models.PL2()
#         resp = pl2.simulated_responses(abilities, params)
#         assert resp.shape == (2, 3)
#         assert np.all(resp >= 0)
#         assert np.all(resp <= 1)



from model import Model
import numpy as np


class PL2(Model):
    """
    2-parameter logistic model (slope/intercept form).
    p(theta) = 1 / (1 + exp(-(zeta + lambda*theta))
    """

    def __init__(self):
        super().__init__(
            PL2.pl2_p,
            num_params=2,
            multi=True,
            prov_params=[0, 1],
            loc_param=0,
            param_bounds=((-10, 10), (0, 5)),
            loc_param_asc=False,
        )

    @staticmethod
    def pl2_p(ability: np.ndarray, params: np.ndarray) -> np.ndarray:
        """
        Probability of a correct response given the ability and item
        parameters.

        Parameters
        ----------
            ability : np.ndarray
                Array of abilities.
            params : np.ndarray
                Item parameters. Item parameters. Each column is a different
                item, each row is a different parameter.

        Returns
        -------
            np.ndarray
                Array of probabilities. Each row is a different individual,
                each column is a different item.
        """
        items = np.array(params).shape[1]
        num_abilities = 1
        if np.array(ability).shape != ():
            num_abilities = np.array(ability).shape[0]
        ability = np.tile(np.array([ability]).transpose(), (1, items))

        params = np.tile(params, num_abilities).reshape((2, num_abilities, items))
        return 1 / (1 + np.exp(-params[0] - (params[1] * ability)))

    @staticmethod
    def convert_param_form(self, params):
        """
        Convert item parameters from one form to another. (Slope/intercept to
        alpha/beta, and vice versa)

        Args:
            params: A list or numpy array representing the item parameters.

        Returns:
            A numpy array representing the converted item parameters.
        """
        return np.array([-params[0] / params[1], 1 / params[1]])



import numpy as np
import matplotlib.pyplot as plt
import matplotlib

class Model:
    """
    Class for item response theory models, e.g. Pl1.
    """

    def __init__(
        self,
        p: callable,
        num_params: int,
        multi: bool = False,
        prov_params: list = None,
        loc_param: int = None,
        param_bounds: list = None,
        loc_param_asc: bool = True,
    ) -> None:
        """
        Initialize model.

        Parameters
        ----------
            p : callable
                Function that calculates the probability of a correct response
                given the ability and item parameters. For example,
                see pl1_p in pl1.py.
            num_params : int
                Number of item parameters. For example the PL3 models uses 3
                parameters: a, b, and c.
            multi : bool, optional
                True if p accepts an array of parameters, false if it accepts a
                single ability and parameters for a single item. Defaults to
                False.
            prov_params : list, optional
                Provisional item parameters, used when EM algorithm is
                initiated. Defaults to None.
            loc_param : int, optional
                Index of location parameter. For example, for the PL3 model
                with item parameters (a,b,c),
                b is the location parameter, so loc_param=1. Defaults to None.
            param_bounds : list, optional
                Minimum and maximum values for each item, for example
                [(0,1), (-5, 5)] for some model with num_params=2.
                Defaults to None.
            loc_param_asc : bool, optional
                True if a higher value for the location parameter results in an
                overall smaller probability of success. Defaults to True.
        """
        if multi:
            self.p = p
        else:
            self.p = lambda abilities, params: np.array(
                [[p(ability, param) for param in params.T] for ability in abilities]
            )
        self.num_params = num_params
        self.prov_params = prov_params
        self.loc_param = loc_param
        self.param_bounds = param_bounds
        self.loc_param_asc = loc_param_asc

    def plot_icc(
        self,
        params: np.ndarray,
    ):
        """
        Plot item characteristic curve for each item.

        Parameters
        ----------
            params : np.ndarray
                Item parameters. Each column is a different item, each row is a
                different parameter.
        """
        ability = np.linspace(-7, 7, 100)
        for i in range(params.shape[1]):
            plt.plot(
                ability,
                self.p(ability, params[:, i].reshape(params.shape[0], 1)),
                label=f"Question{i+1}",
            )
        plt.legend()
        if matplotlib.get_backend().lower() != 'agg':
            plt.show()

    def simulated_responses(
        self,
        abilities: np.ndarray,
        params: np.ndarray,
    ) -> np.ndarray:
        """Randomly generate responses given abilities and item parameters.

        Parameters
        ----------
            abilities (np.ndarray): Array of abilities.
            params (np.ndarray): Item parameters. Each column is a different item, each row is a different parameter.

        Returns
        -------
            np.ndarray
                Array of responses. Each row is a different subject, each column is a different item.
        """
        prob = self.p(abilities, params)
        return (np.random.rand(prob.shape[0], prob.shape[1]) < prob).astype(int)



import numpy as np


def estimate_ability_max_lik(
    self,
    params: np.ndarray,
    results: np.ndarray,
    end=0.00000001,
    eps=0.01,
) -> np.ndarray:
    """
    Estimate abilities using maximum likelihood estimation.
    """
    est = 0.5
    prev_est = 0
    while abs(est - prev_est) > end:
        P = self.prob(est, params)[0]
        W = (1 - P) * P
        h = self.norm_density(est, params)[0]
        prev_est = est
        denom = np.sum(params[1] ** 2 * W)
        if denom < eps or np.any(W < eps):
            break
        est = est + (np.sum(params[1] * W * ((results - P) / h)) / denom)
    return est

def estimate_item_params_max_lik(
    self,
    ability: Union[list, np.ndarray],
    result: Union[list, np.ndarray],
    sigm_orig=0,
    lamb_orig=1,
    end=0.0000001,
    eps=0.1,
) -> np.ndarray:
    """
    Estimate item parameters using maximum likelihood estimation.
    """
    ability = np.array(ability)
    result = np.array(result)
    est = [sigm_orig, lamb_orig]
    prev_est = [0, 0]
    while abs(est[0] - prev_est[0]) > end or abs(est[1] - prev_est[1]) > end:
        P = self.prob(ability, [[est[0]], [est[1]]]).transpose()[0]
        h = self.norm_density(ability, [[est[0]], [est[1]]]).transpose()[0]
        W = h**2 / (P * (1 - P))
        L11 = -np.sum(W)
        L12 = -np.sum(ability * W)
        L22 = -np.sum(ability**2 * W)
        L = np.array([[L11, L12], [L12, L22]])

        if abs(np.linalg.det(L)) < eps:
            break
        L_inv = np.linalg.inv(L)

        v = (result - P) / h
        obs_mat = np.array([np.sum(W * v), np.sum(W * v * ability)])
        prev_est = est
        est = est - np.matmul(L_inv, obs_mat)
    return est



import numpy as np
from item_models import PL1, PL2, PL3
from scipy.stats import norm
import scipy.integrate as integrate
from scipy.optimize import minimize
import matplotlib.pyplot as plt


# mention: we assume there is some sort of location param.
# For graded models we must also assume icc is increasing, so we place bounds
class Mixed:
    def __init__(
        self,
        responses,
        categories,
        models,
        prov_params=None,
        param_bounds=None,
        loc_param_idx=None,
        loc_param_asc=None,
    ) -> None:
        # loc_param asc is a list of booleans, corresponding to each item. The boolean is 'True' if a higher location
        # corresponds to a more difficult item.
        self.responses = responses
        self.num_subjects, self.num_items = responses.shape
        self.categories = categories
        self.models = models
        self.grouped_resp, self.group_idxs, self.group_count = self._group_subjects(
            responses
        )
        self.set_loc_params(loc_param_idx)
        self.ability_distr = lambda ability: norm.pdf(ability)
        self.set_param_bounds(param_bounds)
        self.set_loc_param_asc(loc_param_asc)
        self.set_prov_params(prov_params)

        # print(self.condit_prob_of_indiv_resp(params=self.prov_params, item=0, resp=1, ability=0.4))

    def set_loc_params(self, loc_param):
        # if None, we assume all models either:
        # - are predefined (pl1-pl3, norm)
        # - only have one parameter
        # With this assumption we can construct this object
        # This can be given in the solve step instead of constructor..?
        if loc_param is None:
            self.loc_params = []
            for m in self.models:
                if m.loc_param is not None:
                    self.loc_params.append(m.loc_param)
                else:
                    self.loc_params.append(0)
        else:
            self.loc_params = loc_param

    def set_prov_params(self, prov_params):
        self.prov_params = []
        if prov_params is None:
            for i, m in enumerate(self.models):
                if m.prov_params:
                    self.prov_params.append(
                        np.tile(m.prov_params, (self.categories[i] - 1, 1)).T.astype(
                            float
                        )
                    )

                    # distribute location parameters (assuming bounds exist)
                    lower_bound = self.param_bounds[i][self.loc_params[i]][0]
                    upper_bound = self.param_bounds[i][self.loc_params[i]][1]
                    if self.loc_param_asc[i]:
                        self.prov_params[i][self.loc_params[i]] = np.linspace(
                            lower_bound, upper_bound, self.categories[i] + 1
                        )[1:-1]
                    else:
                        self.prov_params[i][self.loc_params[i]] = np.linspace(
                            upper_bound, lower_bound, self.categories[i] + 1
                        )[1:-1]
                else:
                    raise ValueError('"prov_params" is None')
        else:
            for i, p in enumerate(prov_params):
                self.prov_params.append(
                    np.tile(p, (self.categories[i] - 1, 1)).T.astype(float)
                )

    def set_param_bounds(self, param_bounds):
        # if none, we assume that all items have attribute, or we dont use bounds
        if param_bounds is None:
            self.param_bounds = []
            for m in self.models:
                if m.param_bounds is not None:
                    self.param_bounds.append(m.param_bounds)
                else:
                    self.param_bounds.append([None] * m.num_params)
        else:
            self.param_bounds = param_bounds

    def set_loc_param_asc(self, loc_param_asc):
        if loc_param_asc is None:
            self.loc_param_asc = []
            for m in self.models:
                if m.loc_param_asc is not None:
                    self.loc_param_asc.append(m.loc_param_asc)
                else:
                    self.loc_param_asc.append(True)
        else:
            self.loc_param_asc = loc_param_asc

    def _group_subjects(self, responses):
        return np.unique(responses, axis=0, return_inverse=True, return_counts=True)

    # P*_{i,k}(\theta)
    # 1 when k = -1
    # 0 when k = m-1 ?
    def boundary_prob(self, params, item, resp, ability):
        # resp is single resp
        if resp == -1:
            return 1
        if resp == self.categories[item] - 1:
            return 0
        temp_params = params[item][:, resp].reshape(self.models[item].num_params, 1)
        return self.models[item].p([ability], temp_params)[0][0]

    # P(U_{li}|\theta) = P(U_{li} = k|\theta) = P_{ik}(\theta) = P*_{i,k-1}(\theta) - P*_{ik}(\theta)
    def condit_prob_of_indiv_resp(self, params, resp, item, ability):
        # resp is single resp
        return max(
            self.boundary_prob(params, item, resp - 1, ability)
            - self.boundary_prob(params, item, resp, ability),
            1e-3,
        )

    # P(U_l|\theta) = \prod_{i=1}^n P(U_{li}|\theta)
    def condit_prob_of_resp_vect(self, params, resp, ability):
        return np.prod(
            [
                self.condit_prob_of_indiv_resp(params, u, i, ability)
                for i, u in enumerate(resp)
            ]
        )

    # P(U_l) = \int P(U_l|\theta) \pi(\theta) d\theta
    def marginal_prob_of_resp(self, params, resp):
        integrand = lambda ability: self.condit_prob_of_resp_vect(
            params, resp, ability
        ) * self.ability_distr(ability)
        result, _ = integrate.quad(integrand, -7, 7, epsabs=1e-3, epsrel=1e-3)
        return result

    # L = \sum_{l=1}^{L} r_l \log{P(U_l)}
    def log_likelihood(self, params):
        marg_p = np.array(
            [self.marginal_prob_of_resp(params, resp) for resp in self.grouped_resp]
        )
        if np.any(marg_p <= 0):
            return -np.inf
        return np.sum(self.group_count * np.log(marg_p))

    def loc_param_bounds(self, params, item, param_type_idx, boundary_idx):
        tot_lower_bound = self.param_bounds[item][param_type_idx][0]
        tot_upper_bound = self.param_bounds[item][param_type_idx][1]
        if self.loc_param_asc[item]:
            # lower bound
            if boundary_idx == 0:
                lower = tot_lower_bound
            else:
                lower = params[item][param_type_idx][boundary_idx - 1]
            # upper bound
            if boundary_idx == self.categories[item] - 2:
                upper = tot_upper_bound
            else:
                upper = params[item][param_type_idx][boundary_idx + 1]
        else:
            # lower bound
            if boundary_idx == self.categories[item] - 2:
                lower = tot_lower_bound
            else:
                lower = params[item][param_type_idx][boundary_idx + 1]
            # upper bound
            if boundary_idx == 0:
                upper = tot_upper_bound
            else:
                upper = params[item][param_type_idx][boundary_idx - 1]
        return lower, upper

    def maximize_log_lik_loc(self, params, item, param_type_idx, boundary_idx):
        initial_guess = params[item][param_type_idx][boundary_idx]

        def obj_func(x):
            params[item][param_type_idx][boundary_idx] = x
            return -self.log_likelihood(params)

        result = minimize(
            obj_func,
            initial_guess,
            bounds=[self.loc_param_bounds(params, item, param_type_idx, boundary_idx)],
            tol=1e-3,
        )
        return result.x[0], -result.fun

    def maximize_log_lik(self, params, item, param_type_idx):
        initial_guess = params[item][param_type_idx][0]

        def obj_func(x):
            params[item][param_type_idx] = x
            return -self.log_likelihood(params)

        result = minimize(
            obj_func,
            initial_guess,
            bounds=[self.param_bounds[item][param_type_idx]],
            tol=1e-3,
        )
        return result.x[0], -result.fun

    def em_mmle(self, eps=0.1):
        params = self.prov_params.copy()
        prev_log_L = -np.inf
        log_L = self.log_likelihood(params)
        while abs(log_L - prev_log_L) > eps:
            print(f"prev log L: {prev_log_L}")
            print(f"log L: {log_L}")
            prev_log_L = log_L
            for i in range(len(params)):
                # item_params are all parameters for item i
                for param_type_idx in range(self.models[i].num_params):
                    # param_type are the same type of parameters for i, e.g. all the betas
                    if param_type_idx == self.loc_params[i]:
                        # optimize all location params separately
                        for k in range(self.categories[i] - 1):
                            print(
                                f"maximizing item {i}, {param_type_idx}th parameter, where k={k}"
                            )
                            new_param, log_L = self.maximize_log_lik_loc(
                                params, i, param_type_idx, k
                            )
                    else:
                        print(
                            f"maximizing item {i}, {param_type_idx}th parameter, (all k)"
                        )
                        new_param, log_L = self.maximize_log_lik(
                            params, i, param_type_idx
                        )
        return params

    # \bar(\theta_l) = \frac{\int \theta P(U_l|\theta) \pi(\theta) d\theta}{P(U_l)}
    def estimate_ability_post_mean(self, params, group):
        integrand = (
            lambda ability: self.condit_prob_of_resp_vect(
                params, self.grouped_resp[group], ability
            )
            * self.ability_distr(ability)
            * ability
        )
        numer, _ = integrate.quad(integrand, -7, 7, epsabs=1e-3, epsrel=1e-3)
        denom = self.marginal_prob_of_resp(params, self.grouped_resp[group])
        return numer / denom
    
    def estimate_abilities_post_mean(self, params):
        abilities = np.array([self.estimate_ability_post_mean(params, group) for group in range(len(self.grouped_resp))])
        abilities = abilities[self.group_idxs]
        return abilities

    def plot_boundaries(self, params, i):
        ability = np.linspace(-5, 5)
        for k, param in enumerate(params[i].T):
            prob = self.models[i].p(ability, param.reshape(self.models[i].num_params, 1))
            plt.plot(ability, prob, label=f"P*_{k}")
        plt.legend()
        plt.show()

    def plot_irccc(self, params, i):
        ability = np.linspace(-5, 5)
        for k in range(self.categories[i]):
            prob = [self.condit_prob_of_indiv_resp(params, k, i, ab) for ab in ability]
            plt.plot(ability, prob, label=f"P_{k}")
        plt.legend()
        plt.show()
